#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <semaphore.h>

#include "colli.h"

FILE *fd;
pthread_t interpreterThread;

sem_t *canRead;
sem_t *canSend;

int loop = 1;	
	
void *interpreter(void *cookie){
		
	system("./lassie -s &");
		
	return NULL;
}

ColliStack *colliInit(){

	ColliStack *stack;

	char *sink = "/tmp/colliPipe";
    mkfifo(sink, 0666);
    
	canRead = sem_open("/canRead", O_CREAT, 0644, 0);
 	canSend = sem_open("/canSend", O_CREAT, 0644, 0);
	
	if(canRead == SEM_FAILED || canSend == SEM_FAILED)
		printf("API: %s\n", strerror(errno));
	
	stack = (ColliStack *)malloc(sizeof(ColliStack));
	stack->size = 0;	
	
	pthread_create(&interpreterThread, NULL, interpreter, NULL);	

	return stack;  	
}

int colliPushId(ColliStack *stack, char *id){

	if(!stack || !id)
		return ERR_INVALID_PARAMETER;
		
	if(stack->size == STACKSIZE)
		return ERR_STACK_FULL;

	stack->parameters[stack->size] = strndup(id, strlen(id));
	stack->type[stack->size] = ID;
	stack->size++;
	
	return 0;
}

int colliPushShape(ColliStack *stack, char *shape){

	if(!stack || !shape)
		return ERR_STACK_FULL;
		
	if(stack->size == STACKSIZE)
		return ERR_STACK_FULL;

	stack->parameters[stack->size] = strndup(shape, strlen(shape));
	stack->type[stack->size] = SHAPE;
	
	stack->size++;
	
	return 0;
}

int colliPushCommand(ColliStack *stack, int command){
	
	if(!stack)
		return ERR_INVALID_PARAMETER;
	
	if(stack->size == STACKSIZE)
		return ERR_STACK_FULL;

	if(command < BIND || command > VISUALIZE)
		return ERR_INVALID_PARAMETER;

	stack->command[stack->size] = command;
	stack->type[stack->size] = COMMAND;
	stack->size++;
	
	return 0;
}

int colliPushValue(ColliStack *stack, double value){

	int intValue;
	
	char convertedValue[20];

	if(!stack)
		return ERR_INVALID_PARAMETER;
	
	if(stack->size == STACKSIZE)
		return ERR_STACK_FULL;
	
	if(value < 0){
		value = -value;
		intValue = value * 100;
		snprintf(convertedValue, 20, "-%d.%d", intValue/100, intValue%100);
	}
	else{
		intValue = value * 100;
		snprintf(convertedValue, 20, "%d.%d", intValue/100, intValue%100);
	}
	
	stack->parameters[stack->size] = strndup(convertedValue, strlen(convertedValue));
	stack->type[stack->size] = VALUE;
	stack->size++;

	return 0;
}

int colliPrintStack(ColliStack *stack){

	int i = 0;
	const char *types[] = {"COMMAND", "VALUE", "LIST", "ID"};

	if(!stack)
		return ERR_INVALID_PARAMETER;
		
	
	for(i = 0; i < stack->size; i++)
		printf("[%d] - %s: %s\n", i, types[stack->type[i]], stack->parameters[i]);

	return 0;
}

int colliRun(ColliStack *stack){

	char *id = NULL;
	int command;
	char *shape = NULL;
	char *value = NULL;

	char *xOffset = NULL;
	char *yOffset = NULL;
	
	char full[1024] = "";
	int i = 0;
	
	sem_wait(canSend);	
	
	if(stack->size == 0)
		return ERR_EMPTY_STACK;
	
	if(stack->type[stack->size - 1] != COMMAND) /* If the parameter on top of the stack is not a Command */
		return ERR_BAD_EXPRESSION;
	
	command = stack->command[stack->size - 1];
	stack->size--;
	
	switch(command){
		case BIND:	
			for(i = stack->size - 1; i >= stack->size - 2; i--){		

				if(stack->type[i] == ID)
					id = stack->parameters[i];
			
				if(stack->type[i] == SHAPE)
					shape = stack->parameters[i];
			}

			if(!id || !shape)
				return ERR_BAD_EXPRESSION;

			snprintf(full, 1024, "%s is bound by %s", id, shape);
	
			free(id);
			free(shape);
		
			stack->size -= 2;

		break;
		
		case ROTATE:
	
			for(i = stack->size - 1; i >= stack->size - 2; i--){		
				if(stack->type[i] == ID)
					id = stack->parameters[i];
			
				if(stack->type[i] == VALUE)
					value = stack->parameters[i];
			}

			if(!id || !value)
				return ERR_BAD_EXPRESSION;
	
			snprintf(full, 1024, "%s is rotated %s", id, value);
	
			free(id);
			free(value);
		
			stack->size -= 2;
			
		break;

		case SCALE:
		
			for(i = stack->size - 1; i >= stack->size - 2; i--){		
				if(stack->type[i] == ID)
					id = stack->parameters[i];
			
				if(stack->type[i] == VALUE)
					value = stack->parameters[i];
			}

			if(!id || !value)
				return ERR_BAD_EXPRESSION;
		
			snprintf(full, 1024, "%s is scaled %s", id, value);
	
			free(id);
			free(value);
		
			stack->size -= 2;
		break;

		case TRANSLATE:

			for(i = stack->size - 1; i >= stack->size - 3; i--){		
			
	/*			printf("parameters[%d] = %s\n", i, stack->parameters[i]);*/
			
				if(stack->type[i] == ID){
					id = stack->parameters[i];
					continue;
				}
			
				if(stack->type[i] == VALUE && xOffset == NULL){
					xOffset = stack->parameters[i];
					continue;
				}
			
				if(stack->type[i] == VALUE && yOffset == NULL){
					yOffset = stack->parameters[i];
					continue;
				}
			}

			if(!id || !xOffset || !yOffset)
				return ERR_BAD_EXPRESSION;
		
			snprintf(full, 1024, "%s is translated %s, %s", id, xOffset, yOffset);
		
			free(id);
			free(xOffset);
			free(yOffset);

			stack->size -= 3;
		break;
	
		case SHOW:
		
			if(stack->type[stack->size - 1] == ID)
					id = stack->parameters[stack->size - 1];
		
			if(!id)
				return ERR_BAD_EXPRESSION;
		
			snprintf(full, 1024, "show %s", id);
		
			free(id);		
			stack->size -= 1;
		break;

	
		case VISUALIZE:
		
			snprintf(full, 1024, "show scene");

		break;
		
		default:
			return ERR_INVALID_PARAMETER;
		break;
	}

	fd = fopen("/tmp/colliPipe", "w+");

/*	printf("%s\n", full);*/
	fprintf(fd, "%s\n", full);

	fclose(fd);
	
	sem_post(canRead);

	if(command == VISUALIZE)
		system("./openVisualizer.sh");

	pthread_yield();
	
	return 0;
}

char **colliTest(char *id, int *totalCollisions) {

	int i = 0;
	char testResult[50][15]; 
	
	if(!id)
		return NULL;

/*	printf("collisions with %s\n", id);*/

	sem_t *done = sem_open("/done", O_CREAT, 0644, 0);

	sem_wait(canSend);	
	
		fd = fopen("/tmp/colliPipe", "w+");
	
/*		printf("sending...\n");*/
		fprintf(fd, "collisions with %s\n", id);
	
		fclose(fd);
		
	sem_post(canRead);
	
	sem_wait(done);
	
	FILE *resultFile = fopen("colliResults", "r");
	
	while (fscanf(resultFile, "%s", testResult[i]) != EOF){
	/*		printf("%s\n", testResult[i]);*/
		i++;
	}
	
	fclose(resultFile);
	
	*totalCollisions = i;

	char **result = (char **)malloc(*totalCollisions*sizeof(char *));

/*	printf("Total Collisions = %d\n", *totalCollisions);*/
	
	for(i = 0; i < *totalCollisions; i++){
		result[i] = strndup(testResult[i], strlen(testResult[i]));
/*		printf("result[%d] = %s\n", i, result[i]);*/
	}
	return result;
}

int colliClose(ColliStack *s){

	if(!s)
		return ERR_INVALID_PARAMETER;

	sem_wait(canSend);

		fd = fopen("/tmp/colliPipe", "w+");

		fprintf(fd, "quit\n");

		fclose(fd);

	sem_post(canRead);

	pthread_join(interpreterThread, NULL);

	free(s);
	
	return 0;
}
