%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <unistd.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h>

#include "treeFunctions.h"
#include "parserDefs.h"
#include "symbolTable.h"

char suppress = 0;

%}

%union
{
	int iVal;
    double dVal;
    ternaryTreeNode* tVal;
}

%token<iVal> BOUNDBY IS SCALED TRANSLATED ROTATED COLLISIONTEST IDENTIFIER BRA KET LISTSEPARATOR VARSEPARATOR EOL SHOW SCENE QUIT

%token<iVal> CIRCLE		100
%token<iVal> SQUARE		200
%token<iVal> ALL		300
%token<iVal> REFERENCE	400

%token<dVal> NUMBER
	 
%type<tVal>	script statement assignmentstmt identifier boxstmt optional options scalestmt translationstmt rotationstmt
			collisionstmt referenceshift idList coordList idListElems coordListElems coordinate number newline printstmt visualizestmt
			quitstmt

%%

script : statement 
		{	
//			ternaryTreeNode* node;
//			node = createNode(NOTHING, NOTHING,SCRIPT, $1, NULL, NULL);
//					
//			if(!node)
//				return;
//							
//			printTree(node, 0);
		}
		
		| statement newline script
		{				
//			ternaryTreeNode* node;
//			node = createNode(NOTHING, NOTHING, SCRIPT, $1, $2, NULL);
//				
//			if(!node)
//				return;
//			
////			printTree(node, 0);
//		
		}	
	;

statement : /* empty */
		{
			$$ = createNode(NOTHING, NOTHING, NOTHING, NULL, NULL, NULL);
		
			run($$);	
		}
		| assignmentstmt
		{
			$$ = createNode(NOTHING, NOTHING, ASSIGNMENTSTMT, $1, NULL, NULL);
			
			run($$);
		}
		
		| collisionstmt
		{
			$$ = createNode(NOTHING, NOTHING, COLLISIONSTMT, $1, NULL, NULL);
			
			run($$);
		
		}
				
		| referenceshift
		{
			$$ = createNode(NOTHING, NOTHING, REFERENCESHIFT, $1, NULL, NULL);
		
			run($$);
		}
		| printstmt
		{
			$$ = createNode(NOTHING, NOTHING, PRINTSTMT, $1, NULL, NULL);
		
			run($$);
		}
		| visualizestmt
		{		
			$$ = createNode(NOTHING, NOTHING, VISUALIZESTMT, $1, NULL, NULL);
			
			run($$);
		}
		| quitstmt
		{
			//cleanUp();
		
			return 0;
		}
		| error
		{
			yyerrok;
			yyclearin;
			
			if(!suppress)
				printf(">> ");
		}
	;

assignmentstmt: identifier IS boxstmt
		{
			$$ = createNode(NOTHING, NOTHING, ASSIGNMENTSTMT, $1, $3, NULL);			
		}
		| identifier IS scalestmt
		{
			$$ = createNode(NOTHING, NOTHING, ASSIGNMENTSTMT, $1, $3, NULL);
		}
		| identifier IS translationstmt
		{		
			$$ = createNode(NOTHING, NOTHING, ASSIGNMENTSTMT, $1, $3, NULL);
		}
		| identifier IS rotationstmt
		{		
			$$ = createNode(NOTHING, NOTHING, ASSIGNMENTSTMT, $1, $3, NULL);
		}
	;

boxstmt: BOUNDBY identifier optional
		{
			$$ = createNode(ID, NOTHING, BOXSTMT, $2, $3, NULL);
		}
		| BOUNDBY coordList optional
		{
			$$ = createNode(COORDLIST, NOTHING, BOXSTMT, $2, $3, NULL);
		}
		| BOUNDBY SQUARE optional
		{
			$$ = createNode(SQUARE, NOTHING, BOXSTMT, $3, NULL, NULL);
		}
		| BOUNDBY CIRCLE optional
		{
			$$ = createNode(CIRCLE, NOTHING, BOXSTMT, $3, NULL, NULL);
		}
	;

optional: /* empty */
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONAL, NULL, NULL, NULL);
		} 
		| options
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONAL, $1, NULL, NULL);
		}
		| options options
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONAL, $1, $2, NULL);
		}
		| options options options
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONAL, $1, $2, $3);
		}
	;

options: translationstmt
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONS, $1, NULL, NULL);
		}
		|rotationstmt
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONS, $1, NULL, NULL);
		}
		|scalestmt
		{
			$$ = createNode(NOTHING, NOTHING, OPTIONS, $1, NULL, NULL);
		}
	;

translationstmt: TRANSLATED coordinate
		{
			//Repeated info on iVal so that run() knows what to do...
			$$ = createNode(TRANSLATIONSTMT, NOTHING, TRANSLATIONSTMT, $2, NULL, NULL);
		}
	;

rotationstmt: ROTATED number
		{
			//Repeated info on iVal so that run() knows what to do...
			$$ = createNode(ROTATIONSTMT, NOTHING, ROTATIONSTMT, $2, NULL, NULL);
		}
	;

scalestmt: SCALED number
		{
			//Repeated info on iVal so that run() knows what to do...
			$$ = createNode(SCALESTMT, NOTHING, SCALESTMT, $2, NULL, NULL);
		}
	;

collisionstmt: COLLISIONTEST identifier
		{
			$$ = createNode(ID, NOTHING, COLLISIONSTMT, $2, NULL, NULL);
		}
		
		| COLLISIONTEST idList
		{
			$$ = createNode(NOTHING, NOTHING, COLLISIONSTMT, $2, NULL, NULL);
		}
		
		| COLLISIONTEST ALL
		{
			$$ = createNode(ALL, NOTHING, COLLISIONSTMT, NULL, NULL, NULL);
		}
		
	;

referenceshift: REFERENCE IS coordinate
		{
			$$ = createNode(NOTHING, NOTHING, REFERENCESHIFT, $3 , NULL, NULL);
		}
	;

printstmt: SHOW identifier
		{
			$$ = createNode(NOTHING, NOTHING, SHOW, $2, NULL, NULL);
		}
	;

visualizestmt: SHOW SCENE
		{
			$$ = createNode(NOTHING, NOTHING, VISUALIZESTMT, NULL, NULL, NULL);
		}
	;

idList : BRA idListElems KET
		{
			$$ = createNode(NOTHING, NOTHING, IDLIST, $2, NULL, NULL);
		}
	;

idListElems : identifier
		{
			$$ = createNode(NOTHING, NOTHING,IDLISTELEMS, $1, NULL, NULL);
		}
		
		| idListElems LISTSEPARATOR identifier
		{
			$$ = createNode(NOTHING, NOTHING, IDLISTELEMS, $1, $3, NULL);
		}
	;

coordList : BRA coordListElems KET
		{
			$$ = createNode(COORDLIST, NOTHING, COORDLIST, $2, NULL, NULL);
		}
	;

coordListElems : coordinate
		{
			$$ = createNode(NOTHING, NOTHING, COORDLISTELEMS, $1, NULL, NULL);
		}
		
		| coordListElems LISTSEPARATOR coordinate
		{
			$$ = createNode(NOTHING, NOTHING, COORDLISTELEMS, $1, $3, NULL);
		}
	;

identifier: IDENTIFIER
		{
//			printTable(symbolTable);			
		
			$$ = createNode($1, NOTHING, ID, NULL, NULL, NULL);
		}
	;

coordinate: number VARSEPARATOR number
		{		
			$$ = createNode(NOTHING, NOTHING, COORDINATE, $1, $3, NULL);	
		}
	;
	
number: NUMBER
	{
		$$ = createNode(NOTHING, $1, NUM, NULL, NULL, NULL);	
		
	}
;

newline: EOL
	{
		if(isatty(STDIN_FILENO))
			printf(">> ");
	}
;

quitstmt: QUIT
	{
		cleanUp();
	}
;
	

%%		 

void *parseInit(void *cookie){
	
	yyparse();
		
	return NULL;
}

int main(int argc, char* argv[]) { 

	extern FILE *yyin;
	
	char *semRead = "/canRead";
	char *semSend = "/canSend";
	
	pthread_t parserThread;
	
	sem_t *canRead;
	sem_t *canSend;
	
//	printf("argc = %d\n", argc);

	switch(argc){
	
		case 1:
			yyin = stdin;
		break;
		
		case 2:

			suppress = 1;
	
			if(strcmp(argv[1], "-s") == 0){
						
				char *source = "/tmp/colliPipe";
				mkfifo(source, 0666);
			
				yyin = fopen(source, "w+");
			}
			else{
				printf("Error: unknown option %s\n", argv[1]);
				return -1;
			}
	
			canRead = sem_open(semRead, O_CREAT);
	 		canSend = sem_open(semSend, O_CREAT);
	
		break;
		
		default:
			printf("Error: too many parameters\n");
		break;	
	}
	
	#if YYDEBUG == 1
		extern int yydebug;
		yydebug = 1;
	#endif
	
	initTables();	

	if(isatty(STDIN_FILENO)){
		printf("Colli 0.0.1 - 2014 - Universidade de Brasília\n");
		printf(">> ");
	}
		
	pthread_create(&parserThread, NULL, parseInit, NULL);
	
	pthread_yield();
	
	if(suppress)
		sem_post(canSend);
	
//	yyparse();

	pthread_join(parserThread, NULL);

	return 0; 
}   

int yyerror(char *s, char *id) {
	
	extern char* yytext;
	extern int yylineno;
	
	if(!id)
    	fprintf(stderr, "Error at (%d) \"%s\". %s\n", yylineno, yytext, s); 
	else
		fprintf(stderr, "Error at (%d). Unknown identifier \"%s\"\n.", yylineno, id);

	return 0;
}

#include "lex.yy.c"
