%{

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defs.h"

%}

%union
{
    int iVal;
    double dVal;
    char cVal;
	TERNARY_TREE tVal;
}

%token<iVal> CODE DECLARATIONS OF TYPE CHARACTER INTEGER REAL ENDP IF THEN
			 ELSE ENDIF WHILE ENDWHILE DO ENDDO FOR IS BY TO ENDFOR WRITE NEWLINE
			 READ NOT AND OR IDENTIFIER C_CONST I_CONST R_CONST COLON 
			 SEMICOLON COMMA BRA KET FULLSTOP EQ UNEQ LT GT LEQ GEQ ASSIGN 
			 PLUS MINUS MULT DIV
			 
%type<tVal>  program block identifier_list declaration_block type 	
			 statement_list statement assignment_statement if_statement
			 do_statement while_statement for_statement for_control write_statement
			 output_list read_statement condition conditional comparator expression
			 term value constant identifier

%%

program : identifier COLON block ENDP identifier FULLSTOP
			{	
				if($1->item != $5->item){
					yyerror("Program name mismatch.");
					return;
				}
				
				TERNARY_TREE ParseTree;
				ParseTree = create_node(NOTHING,PROGRAM,$1,$3,$5);
				
				if(!ParseTree)
					return;
				
				#ifdef DEBUG	
					printTree(ParseTree, 0);
				#else
					writeCode(ParseTree, 0);
				#endif
			}
		;
		
block : CODE statement_list 
			{
				$$ = create_node(NOTHING,BLOCK,$2,NULL,NULL);
			}
	  | DECLARATIONS declaration_block CODE statement_list 
			{   			
				$$ = create_node(NOTHING,BLOCK,$2,$4,NULL);			
			}
	  ;
	  
identifier_list : identifier  
					{
						$$ = create_node(NOTHING,IDENTIFIER_LIST,$1,NULL,NULL);
					
						if(check(VERIFY, IDENTIFIER_LIST)){
   							yyerror("Redefinition of identifier.");
   							return;
   						}
			
					}
				| identifier_list COMMA identifier 
					{
						$$ = create_node(NOTHING,IDENTIFIER_LIST,$1,$3,NULL);
					
						if(check(VERIFY, IDENTIFIER_LIST)){
   							yyerror("Redefinition of identifier.");
   							return;
   						}					
   					}
				;

declaration_block : identifier_list OF TYPE type SEMICOLON 
					{
						$$ = create_node(NOTHING,DECLARATION_BLOCK,$1,$4,NULL);
					
						varNum = currentVarListSize;
	
					}
				  | declaration_block identifier_list OF TYPE type SEMICOLON 
					{
						$$ = create_node(NOTHING,DECLARATION_BLOCK,$1,$2,$5);
					
						varNum = currentVarListSize;
	
					}
				  ;

type : CHARACTER 
		{
			$$ = create_node($1,VARIABLE_TYPE,NULL,NULL,NULL);
		} 
     | INTEGER 
		{
			$$ = create_node($1,VARIABLE_TYPE,NULL,NULL,NULL);
		}
     | REAL
		{
			$$ = create_node($1,VARIABLE_TYPE,NULL,NULL,NULL);
		}
     ;

statement_list : statement					 
					{
						$$ = create_node(NOTHING,STATEMENT_LIST,$1,NULL,NULL);
					} 
		 	   | statement_list SEMICOLON statement 
					{			
						$$ = create_node(NOTHING,STATEMENT_LIST,$1,$3,NULL);
					}
			   ;
			   
statement : assignment_statement
			{
				$$ = create_node(NOTHING,ASSIGNMENT_STMT,$1,NULL,NULL);
			} 
		  | if_statement
		  	{
				$$ = create_node(NOTHING,IF_STMT,$1,NULL,NULL);
			} 
		  | do_statement
		  	{
				$$ = create_node(NOTHING,DO_STMT,$1,NULL,NULL);
			} 
		  | while_statement
		  	{
				$$ = create_node(NOTHING,WHILE_STMT,$1,NULL,NULL);
			}  
		  | for_statement
		  	{
				$$ = create_node(NOTHING,FOR_STMT,$1,NULL,NULL);
			}  
		  | write_statement
		  	{
				$$ = create_node(NOTHING,WRITE_STMT,$1,NULL,NULL);
			}  
		  | read_statement
		  	{
				$$ = create_node(NOTHING,READ_STMT,$1,NULL,NULL);
			} 
		  ;
		  
assignment_statement : expression ASSIGN identifier
						{
							$$ = create_node(NOTHING,ASSIGNMENT_STATEMENT,$1,$3,NULL);
		
							if(currentVarListSize > varNum){
								yyerror("Unknown Identifier");
								return;
							}
						} 
					 ;
					 
if_statement : IF conditional THEN statement_list ENDIF 
				{
					$$ = create_node(NOTHING,IF_STATEMENT,$2,$4,NULL);
				}
	       	 | IF conditional THEN statement_list ELSE statement_list ENDIF
	       	 	{
					$$ = create_node(NOTHING,IF_ELSE,$2,$4,$6);
				}
			 ;

do_statement : DO statement_list WHILE conditional ENDDO
				{
					$$ = create_node(NOTHING,DO_STATEMENT,$2,$4,NULL);
				}
			 ;
			 
while_statement : WHILE conditional DO statement_list ENDWHILE
				{
					$$ = create_node(NOTHING,WHILE_STATEMENT,$2,$4,NULL);
				}
			;
				
for_statement : FOR identifier for_control DO statement_list ENDFOR
				{	
					$$ = create_node(NOTHING,FOR_STATEMENT,$2,$3,$5);
			
					if(currentVarListSize > varNum){
						yyerror("Unknown Identifier");
						return;
					}
				}
		      ;

for_control : IS expression BY expression TO expression
			  {
			  		$$ = create_node(NOTHING, FOR_CONTROL, $2, $4, $6);
			  
			  		if(currentVarListSize > varNum){
						yyerror("Unknown Identifier");
						return;
					}
			  }
		     ;
		      
write_statement : WRITE BRA output_list KET
				{
					$$ = create_node(NOTHING,WRITE_STATEMENT,$3,NULL,NULL);
				} 
				| NEWLINE
				{
					$$ = create_node(NEWLINE,WRITE_STATEMENT,NULL,NULL,NULL);
				}
				;
				
output_list : value
				{	
				
					$$ = create_node(NOTHING,OUTPUT_LIST,$1,NULL,NULL);
								
					if(currentVarListSize > varNum){
						yyerror("Unknown Identifier");
						return;
					}
					
				} 
			| output_list COMMA value
				{
					
					$$ = create_node(NOTHING,OUTPUT_LIST,$1,$3,NULL);

					if(currentVarListSize > varNum){
						yyerror("Unknown Identifier");
						return;
					}
				} 
			;

read_statement : READ BRA identifier KET
				{
					$$ = create_node(NOTHING,READ_STATEMENT,$3,NULL,NULL);
				
					if(currentVarListSize > varNum){
						yyerror("Unknown Identifier");
						return;
					}
				}
			   ;

condition : expression comparator expression 
		  	{
				$$ = create_node(NOTHING,CONDITION,$1,$2,$3);
			}
		  | NOT expression comparator expression
			{
				$$ = create_node(NOT,CONDITION,$2,$3,$4);
			}
 		  ;
 		  
conditional	: condition AND conditional 
			{
				$$ = create_node(AND,CONDITIONAL,$1,$3,NULL);
			}
			| condition OR conditional
			{
				$$ = create_node(OR,CONDITIONAL,$1,$3,NULL);
			}
			| condition  
			{
				$$ = create_node(NOTHING,CONDITIONAL,$1,NULL,NULL);
			}
			
			;
			
comparator : EQ 
		   	{
				$$ = create_node(EQ,COMPARATOR,NULL,NULL,NULL);
			}
		   | UNEQ 
			{
				$$ = create_node(UNEQ,COMPARATOR,NULL,NULL,NULL);
			}
		   | LT 
		   	{
				$$ = create_node(LT,COMPARATOR,NULL,NULL,NULL);
			}
		   | GT 
			{
				$$ = create_node(GT,COMPARATOR,NULL,NULL,NULL);
			}
		   | LEQ 
		   	{
				$$ = create_node(LEQ,COMPARATOR,NULL,NULL,NULL);
			}
		   | GEQ
		   	{
				$$ = create_node(GEQ,COMPARATOR,NULL,NULL,NULL);
			}
		   ;
		   
expression : expression PLUS term
		   	{
				$$ = create_node(PLUS,BINARYOP,$1,$3,NULL);
			} 
		   | expression MINUS term
		   	{
				$$ = create_node(MINUS,BINARYOP,$1,$3,NULL);
			}
		   | term
			{
				$$ = create_node(NOTHING,EXPRESSION,$1,NULL,NULL);
			}
		   ;
		   
term : 	term MULT value 
		{
			$$ = create_node(MULT,BINARYOP,$1,$3,NULL);
		}
	  
	  | term DIV value
		{
			$$ = create_node(DIV,BINARYOP,$1,$3,NULL);
		}	
	  
	  | value
		{
			$$ = create_node(NOTHING,TERM, $1,NULL,NULL);
		
			if(currentVarListSize > varNum){
				yyerror("Unknown Identifier");
				return;
			}
		}
	   ;
	 	
value : identifier 
		{
			$$ = create_node(IDENTIFIER,VALUE,$1,NULL,NULL);
		}
	  | constant 
		{
			$$ = create_node(CONSTANT,VALUE,$1,NULL,NULL);
		}
	  | BRA expression KET
		{
			$$ = create_node(BRA,VALUE,$2,NULL,NULL);
		}
	  ;
	  
constant : I_CONST 
			{
				$$ = create_node($1,INT_CONSTANT,NULL,NULL,NULL);
			}
		 | R_CONST 
			{
				$$ = create_node($1,REAL_CONSTANT,NULL,NULL,NULL);
			}
		 | C_CONST
			{
				$$ = create_node($1,CHARACTER_CONSTANT,NULL,NULL,NULL);
			}
		 ;
			  
identifier : IDENTIFIER
		   	{
		   		check($1, VARIABLE_ID);
		   		
		   		$$ = create_node($1,VARIABLE_ID,NULL,NULL,NULL);
			}
		   ;

%%		 

TERNARY_TREE create_node(int ival, int case_identifier, TERNARY_TREE p1, TERNARY_TREE  p2, TERNARY_TREE  p3)
{
    TERNARY_TREE t;
    t = (TERNARY_TREE)malloc(sizeof(TREE_NODE));
    
    if(!t)
    	return NULL;
  
    t->item = ival;
    
    t->nodeIdentifier = case_identifier;
    
    t->first = p1;
    t->second = p2;
    t->third = p3;
    return (t);
}

void writeCode(TERNARY_TREE t, int control){

	static char repeatFlag = FALSE;

	#define OUTPUT stdout
	
	if (t == NULL) 
		return;
	
	switch(t->nodeIdentifier){
   
   		case PROGRAM:
   			fprintf(OUTPUT,"\n#include <stdio.h>\n");
   			fprintf(OUTPUT,"#include <stdlib.h>\n\n");
   			
   			fprintf(OUTPUT, "int main(void) {\n");
   			writeCode(t->second, INDENTOFFSET);
   			fprintf(OUTPUT, "return 0;\n}\n\n");
   		break;
   		
   		case DECLARATION_BLOCK:
   		
   			if(!t->third){
   				writeCode(t->second, 0); //Visit Type Node
   				writeCode(t->first, 0); //Visit Identifier Node
   			}
   			
   			else{
   				writeCode(t->first, 0); //Visit Declaration Block Node
   				writeCode(t->third, 0); //Visit Type Node
   				writeCode(t->second, 0); //Visit Identifier Node
   			}
			   		
   		break;
   		
   		case IDENTIFIER_LIST:
   		
   			if(t->second){
   				writeCode(t->second, 0);
   				fprintf(OUTPUT, ",");
   				writeCode(t->first, 0);
   			}
   			
   			else{
   				writeCode(t->first, 0);
   				fprintf(OUTPUT, ";\n");
   			}
   				
   		break;
   		
   		case VARIABLE_TYPE:
   		 
   			if(!strcmp((char *)symTab[t->item], "INTEGER"))
   				fprintf(OUTPUT, "int ");
   			else if (!strcmp((char *)symTab[t->item], "CHARACTER"))
   				 fprintf(OUTPUT, "char ");
   			else if (!strcmp((char *)symTab[t->item], "REAL"))
   				 fprintf(OUTPUT, "double ");
   				 
   		break;
   		
   		case READ_STATEMENT:
   		
   			fprintf(OUTPUT, "scanf(\"");
   			
   			switch(getType(t->first->item)){
   			
   				case I_TYPE:
   					fprintf(OUTPUT, "%%d");
   				break;
   				
   				case R_TYPE:
   					fprintf(OUTPUT, "%%lf");
   				break;
   				
   				case C_TYPE:
   					fprintf(OUTPUT, "%%c");
   				break;
   				
   				default:
   					fprintf(stderr, "Type Unknown\n"); 
   				break;
   			}
   			
   			fprintf(OUTPUT, "\", &");
   			
   			writeCode(t->first, 0);
   			
   			fprintf(OUTPUT, ");\n");
   		
   		break;
   		
   		case VARIABLE_ID:
  				
   			fprintf(OUTPUT, "%s", (char *)symTab[t->item]);
 
   		break;
   		
   		case WRITE_STATEMENT:	
   			
			if(t->item == NEWLINE)
   				fprintf(OUTPUT, "printf(\"\\n\");\n");
   			else{
   			
				fprintf(OUTPUT, "printf(\"");

   				writeFormat(t->first);

				fprintf(OUTPUT, "\", ");	

   				writeCode(t->first, 0);
				
				fprintf(OUTPUT, ");\n");
				
			}
		break;
		
		case ASSIGNMENT_STATEMENT:
		
			writeCode(t->second, 0);
			
			fprintf(OUTPUT, " = ");
		
			writeCode(t->first, 0);
			
			fprintf(OUTPUT, ";\n");
		
		break;
		
		case FOR_STATEMENT:
		
			fprintf(OUTPUT, "for(");
			
			writeCode(t->first,0);
		
			fprintf(OUTPUT, " = ");
			
			writeCode(t->second->first,0);
			
			fprintf(OUTPUT, "; ");
			
			writeCode(t->first,0);
		
			fprintf(OUTPUT, " <= ");
			
			writeCode(t->second->third,0);
			
			fprintf(OUTPUT, "; ");
			
			writeCode(t->first,0);
		
			fprintf(OUTPUT, " += ");
			
			writeCode(t->second->second,0);
			
			fprintf(OUTPUT, "){\n");
			
			writeCode(t->third, 0);
			
			fprintf(OUTPUT, "}\n");
			
		break;
		
		case DO_STATEMENT:
		
			fprintf(OUTPUT, "do{\n");
			
			writeCode(t->first, 0);
			
			fprintf(OUTPUT, "}while(");
			
			writeCode(t->second, 0);
			
			fprintf(OUTPUT, ");\n");
			
		break;
		
		case WHILE_STATEMENT:
		
			fprintf(OUTPUT, "while(");
			
			writeCode(t->first, 0);
			
			fprintf(OUTPUT, "){\n");
			
			writeCode(t->second, 0);
			
			fprintf(OUTPUT, "};\n");
		
		break;
		
		case IF_STATEMENT:
		
			fprintf(OUTPUT, "if(");
			
			writeCode(t->first, 0);
			
			fprintf(OUTPUT, "){\n");
			
			writeCode(t->second, 0);
			
			fprintf(OUTPUT, "}\n");
		
		break;
		
		case IF_ELSE:
		
			fprintf(OUTPUT, "if(");
			
			writeCode(t->first, 0);
			
			fprintf(OUTPUT, "){\n");
			
			writeCode(t->second, 0);
			
			fprintf(OUTPUT, "}\n");
		
			//ELSE
		
			fprintf(OUTPUT, "else{\n");
				
			writeCode(t->third, 0);
			
			fprintf(OUTPUT, "}\n");
		
		break;
		
		case CONDITIONAL:
		
			if(t->item == NOTHING)
				writeCode(t->first, 0);
			else{
			
				writeCode(t->first, 0);
				
				if(t->item == AND)
					fprintf(OUTPUT, " && ");
				else if(t->item == OR)
					fprintf(OUTPUT, " || ");
					
				writeCode(t->second, 0);
			}
		
		break;
		
		case CONDITION:
		
			if(t->item == NOTHING){
				writeCode(t->first, 0);
				writeCode(t->second, 0);
				writeCode(t->third, 0);
			}
			else if(t->item == NOT){
				
				fprintf(OUTPUT, "!(");
				
				writeCode(t->first, 0);
				writeCode(t->second, 0);
				writeCode(t->third, 0);
				
				fprintf(OUTPUT, ")");
			}
			
		break;
		
		case COMPARATOR:
		
			writeCode(t->first, 0);
		
			switch(t->item){
			
				case EQ:
					fprintf(OUTPUT, " == ");
				break;
		
				case UNEQ:
					fprintf(OUTPUT, " != ");
				break;
		
				case LT:
					fprintf(OUTPUT, " < ");
				break;
			
				case GT:
					fprintf(OUTPUT, " > ");
				break;
				
				case GEQ:
					fprintf(OUTPUT, " >= ");
				break;
		
				case LEQ:
					fprintf(OUTPUT, " <= ");
				break;
		
			}
		
		break;
		
		case BINARYOP:
		
			writeCode(t->first, 0);
			
			switch(t->item){
			
				case PLUS:
				
					fprintf(OUTPUT, " + ");
				
				break;
				
				case MULT:
				
					fprintf(OUTPUT, " * ");
				
				break;
				
				case MINUS:
				
					fprintf(OUTPUT, " - ");
				
				break;
				
				case DIV:
				
					fprintf(OUTPUT, " / ");
				
				break;
			}
			
			writeCode(t->second, 0);
			
		break;

		case OUTPUT_LIST:
		
			writeCode(t->first, 0);
		
			if(t->second)
					fprintf(OUTPUT, ", ");	
				
			writeCode(t->second, 0);
			
		break;
		
		case VALUE:
		
			if(t->item == BRA){
			
				fprintf(OUTPUT,"(");
			
				writeCode(t->first, 0);
				
				fprintf(OUTPUT,")");	
			}
			
			else{
			
				writeCode(t->first, 0);
				writeCode(t->second, 0);
				writeCode(t->third, 0);
			}
		
		break;
		
   		case CHARACTER_CONSTANT:
   		

   			fprintf(OUTPUT, "'");
   		
   			fprintf(OUTPUT, "%c", t->item);
 
 			fprintf(OUTPUT, "'");
   		
   		break;
   		
   		case REAL_CONSTANT:
   			fprintf(OUTPUT, "%s", (char *)symTab[t->item]);
   		break;
   		
   		case INT_CONSTANT:
   			fprintf(OUTPUT, "%d", t->item);
   		break;
   		
   		default:
			writeCode(t->first, 0);
			writeCode(t->second, 0);
			writeCode(t->third, 0);
   		break;	
	}
 
 	return;
}

void printTree(TERNARY_TREE t, int indent)
{
   if (t == NULL) 
   	return;
   
	int i = 0;

	for(i = indent; i >= 0; i--)
		printf(" ");

	switch(t->nodeIdentifier){
   		case REAL_CONSTANT:
   			printf("Item: %s ", (char *)symTab[t->item]);
   		break;
   		
   		case CHARACTER_CONSTANT:
   			printf("Item: %c ", t->item);
   		break;
   		
   		case INT_CONSTANT:
   			printf("Item: %d ", t->item);
   		break;
   		
   		case VARIABLE_ID:
   			printf("ID: %s ", (char *)symTab[t->item]);
   		break;
   		
		case BINARYOP:
			printf("Operator: ");
			
			switch(t->item){
			
				case PLUS:
					printf(" + ");
				break;
				
				case MINUS:
					printf(" - ");
				break;
				
				case MULT:
					printf(" * ");
				break;
				
				case DIV:
					printf(" / ");
				break;
			}
		
		break;
		
   		case COMPARATOR:
   			printf("Operator: ");
			
			switch(t->item){
			
				case EQ:
					printf("= ");
				break;
				
				case UNEQ:
					printf("<> ");
				break;
				
				case LT:
					printf("< ");
				break;
				
				case GT:
					printf("> ");
				break;
				
				case LEQ:
					printf("<= ");
				break;
				
				case GEQ:
					printf(">= ");
				break;
			}
			
   		break;
   }
   
   //printf("Item: %d", t->item);
   
   //printf(" nodeIdentifier: %d\n", t->nodeIdentifier);
   
	printf(" nodeIdentifier: %s\n", getByNumber(t->nodeIdentifier));

	printTree(t->first, 0);
    printTree(t->second, 0);
    printTree(t->third, 0);
   
}

void writeFormat(TERNARY_TREE t){

	static char suppress = FALSE;

	if(!t)
		return;

	switch(t->nodeIdentifier){
	
		case INT_CONSTANT:
			fprintf(stdout, "%%d");
		break;
	
		case REAL_CONSTANT:
			fprintf(stdout, "%%lf");
		break;
	
		case CHARACTER_CONSTANT:
			fprintf(stdout, "%%c");
		break;
	
		case VARIABLE_ID:
		
			if(getType(t->item) == I_TYPE)
				fprintf(stdout, "%%d");
		
			if(getType(t->item) == R_TYPE)
				fprintf(stdout, "%%lf");
		
			if(getType(t->item) == C_TYPE)
				fprintf(stdout, "%%c");
		
		break;
		
		default:
			writeFormat(t->first);
			
			if(t->nodeIdentifier != BINARYOP && t->nodeIdentifier != EXPRESSION)
				writeFormat(t->second);
				
		break;
	}
	
	return;
}

char check(int symTabIndex, int nodeID){

	static char repeatFlag = FALSE;	

	int i;

	if(symTabIndex == VERIFY || nodeID != VARIABLE_ID)
		return repeatFlag;

	for(i = 0; i < currentVarListSize; i++){
	
//		printf("%s,", symTab[varList[i]]);
		if(varList[i] == symTabIndex){
			repeatFlag = TRUE;
			return repeatFlag;
		}
	}	
	
	varList[currentVarListSize] = symTabIndex;
	currentVarListSize++;

	return repeatFlag;
}

const char* getByNumber(int n) {

	if( n >= 0 && n < sizeof(ids))
		return ids[n]; 
	else
		return "INDEX_OUT_OF_BOUNDS";
}

#include "lex.yy.c"
