#ifndef DEFS_H
#define DEFS_H

/* These constants are used later in the code */
#define SYMTABSIZE     50
#define IDLENGTH       15
#define NOTHING        -1
#define INDENTOFFSET    2

/* Constant definitions for type lookup */
#define I_TYPE 1
#define R_TYPE 2
#define C_TYPE 3

#define RESET -1
#define VERIFY -2

enum ParseTreeNodeType { 
	PROGRAM, 	
	BLOCK, 
	IDENTIFIER_LIST, 
	DECLARATION_BLOCK, 
	STATEMENT_LIST, 
	STATEMENT, 
	ASSIGNMENT_STATEMENT, 
	IF_STATEMENT,
	IF_ELSE, 
	DO_STATEMENT,
	WHILE_STATEMENT,
	FOR_STATEMENT,		      
	WRITE_STATEMENT, 
	OUTPUT_LIST,
	READ_STATEMENT,
	CONDITION,
	CONDITIONAL	,
	COMPARATOR,
	EXPRESSION, 
	TERM, 
	VALUE, 
	CONSTANT,  
	CHARACTER_CONSTANT, 
	INT_CONSTANT,  
	REAL_CONSTANT,
	VARIABLE_ID,
	VARIABLE_TYPE,
	FOR_CONTROL,
	BINARYOP,
	ASSIGNMENT_STMT, 
	IF_STMT,
	DO_STMT,
	WHILE_STMT,
	FOR_STMT,		      
	WRITE_STMT, 
	READ_STMT,	
};

const char* ids[] = {"PROGRAM", "BLOCK", "IDENTIFIER_LIST", "DECLARATION_BLOCK", "STATEMENT_LIST", "STATEMENT", "ASSIGNMENT_STATEMENT", "IF_STATEMENT", "IF_ELSE", "DO_STATEMENT", "WHILE_STATEMENT", "FOR_STATEMENT", "WRITE_STATEMENT", "OUTPUT_LIST", "READ_STATEMENT", "CONDITION", "CONDITIONAL", "COMPARATOR", "EXPRESSION",  "TERM", "VALUE", "CONSTANT",  "CHARACTER_CONSTANT", "INT_CONSTANT", "REAL_CONSTANT", "VARIABLE_ID", "VARIABLE_TYPE", "FOR_CONTROL", "BINARYOP", "ASSIGNMENT_STMT", "IF_STMT", "DO_STMT", "WHILE_STMT", "FOR_STMT", "WRITE_STMT", "READ_STMT"};   
/* Add more types here, as more nodes added to tree */

const char *reserved[] ={"CODE","DECLARATIONS","OF","TYPE","CHARACTER","INTEGER","REAL","ENDP","IF","THEN","ELSE","ENDIF","WHILE","ENDWHILE","DO","ENDDO","FOR","IS","BY","TO","ENDFOR","WRITE","NEWLINE","READ","NOT","AND","OR"}; 

#ifndef TRUE
	#define TRUE 1
#endif

#ifndef FALSE
	#define FALSE 0
#endif

#ifndef NULL
	#define NULL 0
#endif

/* ------------- parse tree definition --------------------------- */

struct treeNode {
    int  item;
    int  nodeIdentifier;
    struct treeNode *first;
    struct treeNode *second;
    struct treeNode *third;
  };

typedef  struct treeNode TREE_NODE;
typedef  TREE_NODE        *TERNARY_TREE;

/* ------------- forward declarations --------------------------- */

TERNARY_TREE create_node(int,int,TERNARY_TREE,TERNARY_TREE,TERNARY_TREE);

void printTree(TERNARY_TREE t, int indent);
void writeCode(TERNARY_TREE t, int indent);

const char* getByNumber(int n);
void writeFormat(TERNARY_TREE t);
char check(int index, int nodeID);

extern yyerror(char *s);

//Variables used to verify identifier declaration during compilation
int varList[SYMTABSIZE];
int currentVarListSize = 0;
int varNum = 1;

/* ------------- symbol table definition --------------------------- */

struct symTabNode {
    char identifier[IDLENGTH];
};

typedef  struct symTabNode SYMTABNODE;
typedef  SYMTABNODE        *SYMTABNODEPTR;

SYMTABNODEPTR  symTab[SYMTABSIZE]; 

int currentSymTabSize = 0;

SYMTABNODEPTR newSymTabNode();
int installId(char *id); 
int lookup(char *s);

#endif

