#include "collision.h"
#include "vectorMath.h"

#include <stdlib.h>
#include <stdio.h>

char circleCircle(doubleVector *A, doubleVector *B){

	double dx = A->centerX - B->centerX;
	double dy = A->centerY - B->centerY;
		
	double radii = A->vectorElems[0] + B->vectorElems[0];

	if ((dx * dx) + (dy * dy) < radii * radii)		
		return COLLISION;

	else
		return NO_COLLISION;
}

char polyPoly(doubleVector *A, doubleVector *B){

	int i;
	int j;

	double u_x[2] = {1.0, 0.0};
	double directions[12][2]; //12 different vectors for the test directions

	double *shadows;
 
 	for (i = 0; i < 12; i++){
 		double *rotated = rotate(u_x, i*PI/12);;
 	
 		directions[i][0] = rotated[0];
 		directions[i][1] = rotated[1];
 	
 	}

	for (i = 0; i < 12; i += 1){	   	
 		double u[2];
 		
 		u[0] = directions[i][0];
 		u[1] = directions[i][1];	  	
	   		   	
	   	double pt[2] = {0.0 , 0.0};
	   	
		double *A_coords = (double *)malloc((A->size/2)*sizeof(double));
		double *B_coords = (double *)malloc((B->size/2)*sizeof(double));
		
		for(j = 0; j < A->size; j+=2){
		
			pt[0] = A->vectorElems[j];
			pt[1] = A->vectorElems[j + 1];
		
			A_coords[j/2] = dotProduct(pt, u); //Scalar projection of a vertex onto the test axis
		}
		
		for(j = 0; j < B->size; j+=2){
		
			pt[0] = B->vectorElems[j];
			pt[1] = B->vectorElems[j + 1];
		
			B_coords[j/2] = dotProduct(pt, u); //Scalar projection of a vertex onto the test axis
		}
		
		double Amax = A_coords[0];
		double Amin = A_coords[0];
				
		for(j = 0; j < A->size/2; j++){
		
			if(A_coords[j] > Amax){
				Amax = A_coords[j];
			}
			if(A_coords[j] < Amin){
				Amin = A_coords[j];
			}
		}
		
		double Bmax = B_coords[0];
		double Bmin = B_coords[0];
		
		for(j = 0; j < B->size/2; j++){
				
			if(B_coords[j] > Bmax){
				Bmax = B_coords[j];
			}
			if(B_coords[j] < Bmin){
				Bmin = B_coords[j];
			}
		}
		
		double minmax;
		
		if(Amax < Bmax)
			minmax = Amax;
		else
			minmax = Bmax;

		double maxmin;	

		if(Amin > Bmin)
			maxmin = Amin;
		else
			maxmin = Bmin;

		double shadow = minmax - maxmin;
 
		if (shadow < 0)
			return NO_COLLISION;
	}
	
	return COLLISION;
}

char circlePoly(doubleVector *A, doubleVector *B){

	//Assuming A is Circle, and B is Convex

	int i;
	int j;

	double pt[2] = {0.0, 0.0};
	   		   	
	double *polyProjections = (double *)malloc((B->size/2)*sizeof(double));
		
	pt[0] = A->centerX - B->centerX;
	pt[1] = A->centerY - B->centerY;
 		
	double normalized[2];
	normalized[0] = pt[0]/norm(pt);
	normalized[1] = pt[1]/norm(pt);
		
	double *sizedR = scalarMult(A->vectorElems[0], normalized);
		
	double closestA[2]; //Closest point of A to B
		
	closestA[0] = -sizedR[0] + A->centerX;
	closestA[1] = -sizedR[1] + A->centerY;
	
	for(j = 0; j < B->size; j+=2){
		
		pt[0] = B->vectorElems[j];
		pt[1] = B->vectorElems[j + 1];
		
		polyProjections[j/2] = dotProduct(pt, normalized); //Scalar projection of a vertex onto the normalized center-to-center vector
	}

	double closestB[2];
	
	double Bmax = polyProjections[0];
	
	for(i = 0; i < B->size/2; i++){
		
		if(polyProjections[i] >= Bmax){
			
			Bmax = polyProjections[i];
			
			closestB[0] = B->vectorElems[2*i];
			closestB[1] = B->vectorElems[2*i + 1];
		}
	}

	double centerA[2] = {A->centerX, A->centerY};
	double centerB[2] = {B->centerX, B->centerY};

	double *rA = vectorSub(closestA, centerA);
	double *rB = vectorSub(closestB, centerB);

	char result = COLLISION;

	if(abs(rA[0]) + abs(rB[0]) < abs(centerA[0] - centerB[0]))
		result = NO_COLLISION;
		
	else if(abs(rA[1]) + abs(rB[1]) < abs(centerA[1] - centerB[1]))
		result = NO_COLLISION;
	
	free(rA);
	free(rB);
	free(polyProjections);
	free(sizedR);
			
	return result;
}
