#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <semaphore.h>

#include "treeFunctions.h"
#include "collision.h"

/*
*	This file contains functions pertaining to the execution and evaluation of the parse tree.
*
*/

double referenceX = 0.0;
double referenceY = 0.0;

extern char suppress;

const char* ids[] = {"SCRIPT","STATEMENT","ASSIGNMENTSTMT","ID","BOXSTMT","OPTIONAL",
					"OPTIONS","SCALESTMT","TRANSLATIONSTMT","ROTATIONSTMT","COLLISIONSTMT","REFERENCESHIFT","IDLIST",
					"COORDLIST","IDLISTELEMS","COORDLISTELEMS","COORDINATE", "NUM", "NEWLINE", "PRINT", "SHOW", "QUIT", "VISUALIZE"};  

ternaryTreeNode* createNode(int iVal, double dVal,int nodeType, ternaryTreeNode* p1, ternaryTreeNode*  p2, ternaryTreeNode*  p3)
{
    ternaryTreeNode* t;
    t = (ternaryTreeNode *)malloc(sizeof(ternaryTreeNode));
    
    if(!t){
  		yyerror("Unexpected interpreter error: out of memory.\n", NULL);
    	return NULL;
  	}
    t->iVal = iVal;
    t->dVal = dVal;
    
    t->nodeType = nodeType;
        
    t->first = p1;
    t->second = p2;
    t->third = p3;
    
    return t;
}

void run(ternaryTreeNode* node){
 
 	extern SymbolTable symbolTable[SYMTABSIZE];
 
 	static doubleVector *objects[MAX_OBJECTS];
 	static int currentObjects = 0;
	doubleVector *toPrint = NULL;
	doubleVector *toTest = NULL;
 	int workIndex;

	char *semRead = "/canRead";
	char *semSend = "/canSend";
	
	sem_t *canRead;
	sem_t *canSend;

 	int i, j; //Loop Variables
 
 	int id = 0;
 	int type = 0;
 	ternaryTreeNode* options = NULL;
  	
 	double centerX = 0.0;
 	double centerY = 0.0;

	//Variables used to communicate with the visualizer
	int stream;
	static char *channel;
	
	FILE *result;
	
	if(suppress){
		canRead = sem_open(semRead, O_CREAT);
 		canSend = sem_open(semSend, O_CREAT);
 		
		if(canRead == SEM_FAILED || canSend == SEM_FAILED)
			printf("Interpreter: %s\n", strerror(errno));
 	}
 	 	
 	if(suppress)	
 		sem_wait(canRead);
 	
/*	puts("Doing the thing...");	*/
 	
 	switch(node->nodeType){
  
 		case ASSIGNMENTSTMT:
 		
 			id = node->first->first->iVal;
 			type = node->first->second->iVal;
 			options = node->first->second->first;
 			
/* 			puts("*********OPTIONS*********");*/
/* 				printTree(options, 0 );*/
/* 			puts("*********OPTIONS*********");*/
 			 			
 			switch(type){
	 			
	 			case COORDLIST:
	 				
	 				options = node->first->second->second;
	 				
/*	 				printf("Id under test: %d\n", id);*/
	 				
	 				if(!isAssigned(id)){
	 					assign(id);
	 					workIndex = currentObjects;
	 					currentObjects++;
	 				}
	 				else{
	 					workIndex = getIndex(objects, currentObjects, id);
	 					free(objects[workIndex]);
	 				}

	 				ternaryTreeNode *list = node->first->second->first->first;

					int listSize = checkSize(list, 1);	

					if(listSize/2 == 2) //if the list describes a Line Segment
						yyerror("Can't create a line segment!", NULL);
	 
					objects[workIndex] = createDoubleVector(listSize);
	 				
	 				fillObject(objects[workIndex], list);
					
	 				objects[workIndex]->self = id;
	 				
/*	 				if(isconvex(object[workIndex]))*/
	 					objects[workIndex]->type = CONVEX_SHAPE;	 			
/*					else*/
/*						objects[workIndex]->type = CONCAVE_SHAPE;*/

					double *W = (double *)calloc(listSize/2, sizeof(double));

					if(!W){
						yyerror("Out of memory!", NULL);
						return;
					}

					for(i = 0; i < listSize; i+=2){
					
						double x0 = objects[workIndex]->vectorElems[i];
						double y0 = objects[workIndex]->vectorElems[i + 1];
						
						double x1 = objects[workIndex]->vectorElems[(i + 2) % listSize];
						double y1 = objects[workIndex]->vectorElems[((i + 2) % listSize) + 1];

						W[i/2] = (0.5 * (y1 * x0 - y0 * x1));	 				
	 				}
	 				
	 				double area = 0;
	 				
	 				for(i = 0; i < listSize/2; i++)
	 					area += W[i];
	 				
	 				
	 				for(i = 0; i < listSize; i+=2){
						
						double x0 = objects[workIndex]->vectorElems[i];
						double y0 = objects[workIndex]->vectorElems[i + 1];
						
						double x1 = objects[workIndex]->vectorElems[(i + 2) % listSize];
						double y1 = objects[workIndex]->vectorElems[((i + 2) % listSize) + 1];
						
						centerX += (x1 + x0) * W[i/2]/3.0;
						centerY += (y1 + y0) * W[i/2]/3.0;
	 				}
	 				
	 				centerX /= area;
	 				centerY /= area;
	 				
	 				objects[workIndex]->centerX = centerX;
					objects[workIndex]->centerY = centerY;
	 			
	 				free(W);
	 			
/*	 				//printTree(options, 0);*/

					if(options){
						applyOption(objects[workIndex], options->first);
			 			applyOption(objects[workIndex], options->second);
			 			applyOption(objects[workIndex], options->third);
			 		}
	 				
	 			break;
	 			
	 			case ID:
	 				
	 				options = node->first->second->second;
	 				
	 				if(!isAssigned(id)){
	 					assign(id);
	 					workIndex = currentObjects;
	 					currentObjects++;
	 				}
	 				else{
	 					workIndex = getIndex(objects, currentObjects, id);
	 					free(objects[workIndex]);
	 				}

	 				doubleVector *toCopy = getById(objects, currentObjects, node->first->second->first->iVal);

					objects[workIndex] = createDoubleVector(toCopy->size);
	 				
	 				for(i = 0; i < toCopy->size; i++)
	 					pushDoubleBack(objects[workIndex], toCopy->vectorElems[i]);
	 				
	 				objects[workIndex]->self = id;
	 				
/*	 				if(isconvex(object[workIndex]))*/
	 					objects[workIndex]->type = toCopy->type;	 			
/*					else*/
/*						objects[workIndex]->type = CONCAVE_SHAPE;*/

	 				objects[workIndex]->centerX = toCopy->centerX;
					objects[workIndex]->centerY = toCopy->centerY;
	 			
/*	 				//printTree(options, 0);*/

					if(options){
						applyOption(objects[workIndex], options->first);
			 			applyOption(objects[workIndex], options->second);
			 			applyOption(objects[workIndex], options->third);
			 		}
	 				
	 			break;
	 			
	 			case TRANSLATIONSTMT:
							 			
					if(!isAssigned(id))
						yyerror("Unknown identifier", symbolTable[id].identifier);
	 						
					makeTranslation(getById(objects, currentObjects, id), node->first->second->first->first->dVal, 
																		  node->first->second->first->second->dVal); 

	 			break;
	 		
	 			case SCALESTMT:
					
					if(!isAssigned(id))
	 					yyerror("Unknown identifier", symbolTable[id].identifier);
					
					makeScaling(getById(objects, currentObjects, id), node->first->second->first->dVal); 

	 			break;

	 			case ROTATIONSTMT:
					
					if(!isAssigned(id))
	 					yyerror("Unknown identifier", symbolTable[id].identifier);
					
					makeRotation(getById(objects, currentObjects, id), node->first->second->first->dVal); 
																		  
	 			break;

	 			case SQUARE_SHAPE:
	 				 				
/*	 				printf("Id under test: %d\n", id);*/
	 				
	 				if(!isAssigned(id)){
	 					assign(id);
	 					workIndex = currentObjects;
	 					currentObjects++;
	 				}
	 				else{
	 					workIndex = getIndex(objects, currentObjects, id);
	 					free(objects[workIndex]);
	 				}
	 				
	 				objects[workIndex] = createDoubleVector(8);
	 				objects[workIndex]->self = id;
	 				objects[workIndex]->type = CONVEX_SHAPE;
	 			
	 				objects[workIndex]->centerX = 0.0  + referenceX;
					objects[workIndex]->centerY = 0.0  + referenceY;
	 			
	 				//printTree(options, 0);

					//Initially creates a square with side equal to 1
					pushDoubleBack(objects[workIndex], 0.5 + referenceX);
					pushDoubleBack(objects[workIndex], 0.5 + referenceY);
			
					pushDoubleBack(objects[workIndex], -0.5 + referenceX);
					pushDoubleBack(objects[workIndex], 0.5 + referenceY);
			
					pushDoubleBack(objects[workIndex], -0.5 + referenceX);
					pushDoubleBack(objects[workIndex], -0.5 + referenceY);
					
					pushDoubleBack(objects[workIndex], 0.5 + referenceX);
					pushDoubleBack(objects[workIndex], -0.5 + referenceY);				
	 		
					if(options){
						applyOption(objects[workIndex], options->first);
			 			applyOption(objects[workIndex], options->second);
			 			applyOption(objects[workIndex], options->third);
			 		}
	 	 					
	 			break;
	 		
	 			case CIRCLE_SHAPE:
	 				
/*	 				printf("Id under test: %d\n", id);*/
	 				
	 				if(!isAssigned(id)){
	 					assign(id);
	 					workIndex = currentObjects;
	 					currentObjects++;
	 				}
	 				else{
	 					workIndex = getIndex(objects, currentObjects, id);
	 					free(objects[workIndex]);
	 				}
	 				objects[workIndex] = createDoubleVector(2);
	 				objects[workIndex]->self = id;
	 				objects[workIndex]->type = CIRCLE_SHAPE;
	 			
	 				objects[workIndex]->centerX = 0.0 + referenceX;
					objects[workIndex]->centerY = 0.0 + referenceY;
					
	 				double radius = 1.0;
	 			
	 				//printTree(options, 0);

					pushDoubleBack(objects[workIndex], radius);
	 				pushDoubleBack(objects[workIndex], radius);

					if(options){
						applyOption(objects[workIndex], options->first);
			 			applyOption(objects[workIndex], options->second);
			 			applyOption(objects[workIndex], options->third);
			 		}
	 			break;
	 		}

 		break;

		case PRINTSTMT:
		
			id = node->first->first->iVal;
		
			if(!isAssigned(id))
	 			yyerror("Unknown identifier ", symbolTable[id].identifier);
		
			toPrint = getById(objects, currentObjects, id);
			
			if(!toPrint)
				break;
			
			printf("Identifier: %s\n", symbolTable[toPrint->self].identifier);	
				
			printf("Center: %.2f, %.2f\n", toPrint->centerX, toPrint->centerY);
			
			printf("Type: ");
			if(toPrint->type == CIRCLE_SHAPE)
				printf("Circle Shape\n");
			else
				printf("Convex Shape\n");
			
			printf("Size: %d\n", toPrint->size/2);
			printf("Contents:\n\t { ");
				printDoubleVector(toPrint);
			printf("}\n");
		
		break;
		
		case VISUALIZESTMT:
		
			stream = open(channel, O_WRONLY);

			FILE *scene = fopen("description.scene", "w");
		
			fprintf(scene, "%d\n", currentObjects);
		
			for(i = 0; i < currentObjects; i++){
		
				if(objects[i]->size == 2)
					fprintf(scene, "(%.4f, %.4f), ", objects[i]->centerX, objects[i]->centerY);
		
				for(j = 0; j < objects[i]->size; j+=2){
				
					fprintf(scene, "(%.4f, %.4f)", objects[i]->vectorElems[j], objects[i]->vectorElems[j + 1]);
				
					if(j != objects[i]->size - 2)
						fprintf(scene, ",");
					else
						fprintf(scene, "\n");
				}
			}

			fclose(scene);

			if(!suppress)
				system("./openVisualizer.sh");
			
		break;
		
		case REFERENCESHIFT:
			
			referenceX = node->first->first->first->dVal;
			referenceY = node->first->first->second->dVal;
			
			for(i = 0; i < currentObjects; i++)
				makeTranslation(objects[i], -referenceX, -referenceY);			
			
		break;
		
  		case COLLISIONSTMT:
	
/*			printf("Identifier under test: %d\n", id);*/

			if(suppress){
/*				puts("getting results file...");*/
				result = fopen("colliResults", "w");
			}
 			type = node->first->iVal;
 			if(type == ID){

				sem_t *done;
				
				if(suppress)
					done = sem_open("/done", O_CREAT);

				id = node->first->first->iVal;

				toTest = getById(objects, currentObjects, id);

 				for(i = 0; i < currentObjects; i++){
 					if(i == id)
 						continue;
 					
 					//If the center coordinates of the objects to be tested coindice, return collision
 					if((fabs(objects[i]->centerX - toTest->centerX) < 1e-6) && (fabs(objects[i]->centerY - toTest->centerY) < 1e-6)){
				   			if(!suppress)
				   				printf("%s collided with %s\n", symbolTable[toTest->self].identifier, 
				   											symbolTable[objects[i]->self].identifier);
 					
 							else
 								fprintf(result, "%s\n", symbolTable[objects[i]->self].identifier);
 							
 					}
 					else{
	 					if(objects[i]->type == CIRCLE_SHAPE && toTest->type == CIRCLE_SHAPE){
							if (circleCircle(objects[i], toTest) == COLLISION){
						   		if(!suppress)
					   				printf("%s collided with %s\n", symbolTable[toTest->self].identifier, 
					   											symbolTable[objects[i]->self].identifier);
	 					
	 							else{
	 								fprintf(result, "%s\n", symbolTable[objects[i]->self].identifier);
								}
							}
						}

						else if(objects[i]->type == CONVEX_SHAPE && toTest->type == CONVEX_SHAPE){
							if (polyPoly(objects[i], toTest) == COLLISION){
					   			if(!suppress)
					   				printf("%s collided with %s\n", symbolTable[toTest->self].identifier, 
					   											symbolTable[objects[i]->self].identifier);
	 					
	 							else
	 								fprintf(result, "%s\n", symbolTable[objects[i]->self].identifier);
							}			
						}
						
						else if(objects[i]->type == CIRCLE_SHAPE && toTest->type == CONVEX_SHAPE){
							if (circlePoly(objects[i], toTest) == COLLISION){      
					   			if(!suppress)
					   				printf("%s collided with %s\n", symbolTable[toTest->self].identifier, 
					   											symbolTable[objects[i]->self].identifier);
	 					
	 							else
	 								fprintf(result, "%s\n", symbolTable[objects[i]->self].identifier);
							}			
						}
						
						else if(objects[i]->type == CONVEX_SHAPE && toTest->type == CIRCLE_SHAPE){
							if (circlePoly(toTest, objects[i]) == COLLISION){      
					   			if(!suppress)
					   				printf("%s collided with %s\n", symbolTable[toTest->self].identifier, 
					   											symbolTable[objects[i]->self].identifier);
	 					
	 							else
	 								fprintf(result, "%s\n", symbolTable[objects[i]->self].identifier);
							}			
						}
					}
				}
 			
 				if(suppress)
 					sem_post(done);	
 			}
 	
 			if(type == ALL_QUERY){
 	
 				int i = 0; 
 				int j = 0;
 	
 				//Naive Implementation, must refactor for greater speed
 				for(i = 0; i < currentObjects; i++){
 					for(j = i; j < currentObjects; j++){			
 						
 						if(i == j) //Skips collision tests on self
 							continue;
	 					 
	 					//If the centers of the objects coincide, they are colliding					
 						if((fabs(objects[i]->centerX - objects[j]->centerX) < 1e-6) 
 						&& (fabs(objects[i]->centerY - objects[j]->centerY) < 1e-6)){
				   			
				   			printf("%s collided with %s\n", symbolTable[objects[i]->self].identifier, 
				   											symbolTable[objects[j]->self].identifier);
 						}
 							
 						else{
	 						if(objects[i]->type == CIRCLE_SHAPE && objects[j]->type == CIRCLE_SHAPE){
	 							
	 								if(circleCircle(objects[i], objects[j]) == COLLISION)
							   			printf("%s collided with %s\n", symbolTable[objects[i]->self].identifier,
							   											symbolTable[objects[j]->self].identifier);			
	 						}
	 						else if(objects[i]->type == CONVEX_SHAPE &&objects[j]->type == CONVEX_SHAPE){
								if (polyPoly(objects[i], objects[j]) == COLLISION)        
						   			printf("%s collided with %s\n", symbolTable[objects[j]->self].identifier, 
						   											symbolTable[objects[i]->self].identifier);			
							}
							
							else if(objects[i]->type == CIRCLE_SHAPE && objects[j]->type == CONVEX_SHAPE){
								if (circlePoly(objects[i], objects[j]) == COLLISION)      
						   				printf("%s collided with %s\n", symbolTable[objects[j]->self].identifier, 
						   											symbolTable[objects[i]->self].identifier);
		 					}
						
							else if(objects[i]->type == CONVEX_SHAPE && objects[j]->type == CIRCLE_SHAPE){
								if (circlePoly(objects[j], objects[i]) == COLLISION)
						   				printf("%s collided with %s\n", symbolTable[objects[j]->self].identifier, 
						   											symbolTable[objects[i]->self].identifier);
							}
							
 						}
 					}
 				}
 			}
 		
 		if(suppress)
 			fclose(result);
 			
 		break;
 	}

/*	printf("Did the Thing!\n");*/

	if(suppress)
		sem_post(canSend);

 	return;
}

int makeTranslation(doubleVector* object, double xOffset, double yOffset){

	int i;

	if(!object){
		return -1;
	}
	
	if(object->type != CIRCLE_SHAPE){
		for(i = 0; i < object->size; i += 2){
			object->vectorElems[i] += xOffset;
			object->vectorElems[i + 1] += yOffset;
		}
	}
	
	object->centerX += xOffset;
	object->centerY += yOffset;
	
	return 0;
}

int makeRotation(doubleVector* object, double angle){

	int i;
	double currentCenterX;
	double currentCenterY;

	if(!object)
		return -1;
	
	if(object->type == CIRCLE_SHAPE){
		return 0;
	}
	
	currentCenterX = object->centerX;
	currentCenterY = object->centerY;
	
	angle = angle*PI/180;

	if(currentCenterX != 0.0 || currentCenterY != 0.0)
		makeTranslation(object, -currentCenterX, -currentCenterY); // Moves object to origin 

	for(i = 0; i < object->size; i+=2){
		double xCoord = object->vectorElems[i];
		double yCoord = object->vectorElems[i + 1];
			
		object->vectorElems[i] = cos(angle)*xCoord + sin(angle)*yCoord;
		object->vectorElems[i + 1] = -sin(angle)*xCoord + cos(angle)*yCoord;
	}

	if(currentCenterX != 0.0 || currentCenterY != 0.0)
		makeTranslation(object, currentCenterX, currentCenterY); // Moves object back to its original positon 
	
	return 0;
}

int makeScaling(doubleVector* object, double ratio){

	int i;
	double currentCenterX;
	double currentCenterY;

	if(!object)
		return -1;
	
	if(object->type == CIRCLE_SHAPE){
		for(i = 0; i < object->size; i++)
			object->vectorElems[i] *= ratio;
	
		return 0;
	}
	
	currentCenterX = object->centerX;
	currentCenterY = object->centerY;

	if(currentCenterX != 0.0 || currentCenterY != 0.0)
		makeTranslation(object, -currentCenterX, -currentCenterY); // Moves object to origin 

	for(i = 0; i < object->size; i++) // Scales object
		object->vectorElems[i] *= ratio;
	
	if(currentCenterX != 0.0 || currentCenterY != 0.0)
		makeTranslation(object, currentCenterX, currentCenterY); // Moves object back to its original positon 
		
	return 0;
}

int getIndex(doubleVector** objectList, int listSize, int id){

	int i;

	if(id < 0)
		return -1;	

	for(i = 0; i < listSize; i++)
		if(objectList[i]->self == id)
			return i;

	return -1;

}

doubleVector* getById(doubleVector** objectList, int listSize, int id){

	int i;

	if(id < 0)
		return NULL;	

	for(i = 0; i < listSize; i++)
		if(objectList[i]->self == id)
			return objectList[i];

	return NULL;
}

int checkSize(ternaryTreeNode* list, char reset){
	
	static int size = 0;
	
	if(!list)
		return -1;
	
	if(reset)
		size = 0;
	
	size++;
		
	checkSize(list->first, 0);
	checkSize(list->second, 0);	

	return (size + 1)/2;
}

int fillObject(doubleVector *object, const ternaryTreeNode *list){

	if(!list)
		return -1;
	
	if(list->nodeType == COORDINATE){
		pushDoubleBack(object, list->first->dVal + referenceX);
		pushDoubleBack(object, list->second->dVal + referenceY);
	}
	
	fillObject(object, list->first);
	fillObject(object, list->second);
	fillObject(object, list->third);
	
	return 0;
}

int applyOption(doubleVector *object, const ternaryTreeNode* option){
	
	if(!option)
		return -1;

	switch(OPTION_TYPE(option)){
		case TRANSLATIONSTMT:
			makeTranslation(object, 0.0 + option->first->first->first->dVal, 0.0 + option->first->first->second->dVal); 
		break;	
	 							
		case SCALESTMT:
			makeScaling(object, option->first->first->dVal);
		break;
	 						
		case ROTATIONSTMT:
			if(object->type != CIRCLE_SHAPE)
			 makeRotation(object, option->first->first->dVal);
		break;
	}
	
	return 0;
}

void printTree(ternaryTreeNode* node, int indentLevel) //Prints some of the contents of the parse tree
{
	int i = 0;

	if (node == NULL) 
		return;
   
	for(i = 2*indentLevel; i >= 0; i--)
		printf(" ");

	printf("Type: %s\n", ids[node->nodeType]);

	printTree(node->first, indentLevel + 1);
	printTree(node->second, indentLevel + 1);    
	printTree(node->third, indentLevel + 1);
 
	indentLevel++;
 
	return;
}

void cleanUp(){

	
	
	return;
}
