%{

#include "defs.h"

#ifndef PRINT
	#define TOKEN(s) return(s);
#else
	#define TOKEN(s) printf("Lexeme: %s \n -> Token returned: %s\n\n", yytext, #s);
#endif

/* Declare Symbol Table Type and Array as imported types */
extern SYMTABNODEPTR symTab[SYMTABSIZE];
extern int currentSymTabSize;

%}

%option yylineno

/*Definitions*/

delim       	[ \t\n]
ws          	{delim}+

character      	[A-Za-z]
digit       	[0-9]

identifier		{character}({character}|{digit})*
integer			("+"|"-")?{digit}+
real			("+"|"-")?{digit}+"."{digit}+

character_constant	"'"{character}"'"
 
%%
            int k = 0;

{ws}        ; /* white space, do nothing */
 
CODE       		TOKEN(CODE)
DECLARATIONS	TOKEN(DECLARATIONS)
OF				TOKEN(OF)
TYPE			TOKEN(TYPE)
CHARACTER		{
					yylval.iVal = installId(yytext);
					TOKEN(CHARACTER)
				}
				
INTEGER			{
					yylval.iVal = installId(yytext);
					TOKEN(INTEGER)
				}
				
REAL			{
					yylval.iVal = installId(yytext);
					TOKEN(REAL)
				}
				
ENDP			TOKEN(ENDP)
IF				TOKEN(IF)
THEN			TOKEN(THEN)
ELSE			TOKEN(ELSE)
ENDIF			TOKEN(ENDIF)
WHILE			TOKEN(WHILE)
ENDWHILE		TOKEN(ENDWHILE)
DO				TOKEN(DO)
ENDDO			TOKEN(ENDDO)
FOR				TOKEN(FOR)
IS				TOKEN(IS)
BY				TOKEN(BY)
TO				TOKEN(TO)
ENDFOR			TOKEN(ENDFOR)
WRITE			TOKEN(WRITE)
NEWLINE			TOKEN(NEWLINE)
READ			TOKEN(READ)
NOT				TOKEN(NOT)
AND				TOKEN(AND)
OR				TOKEN(OR)

{identifier}			{
							yylval.iVal = installId(yytext);
							TOKEN(IDENTIFIER)
						}
												
{character_constant}	{
							//get second char in yytext (char after first quote)
							yylval.cVal = yytext[1];
							TOKEN(C_CONST)
						}
{integer}				{
							yylval.iVal = atoi(yytext);
							TOKEN(I_CONST)
						}
{real}					{
							yylval.iVal = installId(yytext);
							TOKEN(R_CONST)
						}
	
":"						TOKEN(COLON)
";"						TOKEN(SEMICOLON)
","						TOKEN(COMMA)
"("						TOKEN(BRA)
")"						TOKEN(KET)
"."						TOKEN(FULLSTOP)

"="						TOKEN(EQ)
"<>"					TOKEN(UNEQ)
"<"						TOKEN(LT)
">"						TOKEN(GT)
"<="					TOKEN(LEQ)
">="					TOKEN(GEQ)

"->"					TOKEN(ASSIGN)
"+"						TOKEN(PLUS)
"-"						TOKEN(MINUS)
"*"						TOKEN(MULT)
"/"						TOKEN(DIV)

%%

SYMTABNODEPTR newSymTabNode()
{
    return ((SYMTABNODEPTR)malloc(sizeof(SYMTABNODE)));
}


/* Look up an identifier in the symbol table, if its there return
   its index.  If its not there, put it in the end position,
   as long as the table isn't full, and return its index.
*/

int installId(char *id) 
{
    extern SYMTABNODEPTR symTab[SYMTABSIZE]; 
    extern int currentSymTabSize;
    int index;

    index = lookup(id);
    
    if (index >= 0)
    {
        return (index);
    }
    
    else 
    {
       symTab[currentSymTabSize] = newSymTabNode();
       strcpy(symTab[currentSymTabSize]->identifier,id);
       return(currentSymTabSize++);
    }
}

int lookup(char *s)
{
    extern SYMTABNODEPTR symTab[SYMTABSIZE];
    extern int currentSymTabSize;
    int i;

    for(i = 0; i < currentSymTabSize; i++)
    {
    
        if(strcmp(s,symTab[i]->identifier) == 0)
            return i;
        
    }
    
    return -1;    
}

int getType(int index){

	int i = index;
	
	for(i = index; i < currentSymTabSize; i++){
	
		if(strcmp("INTEGER", (char *)symTab[i]) == 0)
			return I_TYPE;
		if(strcmp("REAL", (char *)symTab[i]) == 0)
			return R_TYPE;
		if(strcmp("CHARACTER", (char *)symTab[i]) == 0)
			return C_TYPE;
		
	}

	return -1;

}

